using System;
using System.CodeDom;
using System.Collections;
using System.CodeDom.Compiler;
using System.Text;

namespace swall
{
	/// <summary>
	/// Dynamically creates a wrapper class to create property methods for custom objects. 
	/// A DataGrid can then access these properties and bind properly.
	///
	/// Author: wall@shawnwall.com
	/// feel free to modify at your own will.
	/// </summary>
	public class WebServiceWrapper
	{
		public static Hashtable cachedTypes = new Hashtable();

		private static Type GetWrapperType(Type typeToWrap)
		{
			if(!cachedTypes.ContainsKey(typeToWrap))
				cachedTypes.Add(typeToWrap,GetNewWrapperType(typeToWrap));
			return (Type)cachedTypes[typeToWrap];
		}

		private static Type GetNewWrapperType(Type typeToWrap)
		{
			CodeTypeDeclaration code = new CodeTypeDeclaration(typeToWrap.Name + "WithProperties");
			code.IsClass = true;

			CodeConstructor constructor = new CodeConstructor();
			constructor.Attributes = MemberAttributes.Public;

			CodeMemberField declaration = new CodeMemberField(typeToWrap, "_" + typeToWrap.Name);
			declaration.Attributes = MemberAttributes.Private;
			code.Members.Add(declaration);

			CodeMemberMethod setInnerObject = new CodeMemberMethod();
			setInnerObject.Name = "SetInnerObject";
			setInnerObject.Parameters.Add(new CodeParameterDeclarationExpression(typeToWrap, typeToWrap.Name));
			setInnerObject.Statements.Add(new CodeAssignStatement(new CodeVariableReferenceExpression("_" + typeToWrap.Name), new CodeVariableReferenceExpression(typeToWrap.Name)));
			setInnerObject.Attributes = MemberAttributes.Public;
			code.Members.Add(setInnerObject);

			CodeMemberMethod getInnerObject = new CodeMemberMethod();
			getInnerObject.Name = "GetInnerObject";
			getInnerObject.ReturnType = new CodeTypeReference(typeToWrap);
			getInnerObject.Statements.Add(new CodeMethodReturnStatement(new CodeVariableReferenceExpression("_" + typeToWrap.Name)));
			getInnerObject.Attributes = MemberAttributes.Public;
			code.Members.Add(getInnerObject);

			System.Reflection.FieldInfo fieldinfo;
			System.Reflection.FieldInfo[] fields;
			fields = typeToWrap.GetFields();

			for(int i=0; i < fields.Length; i++)
			{
				fieldinfo = fields[i];
				CodeMemberProperty prop = new CodeMemberProperty();
				prop.Name = fieldinfo.Name;
				prop.Attributes = MemberAttributes.Public;
				prop.HasGet = true;
				prop.HasSet = true;
				prop.Type = new CodeTypeReference(fieldinfo.FieldType);

				//create getter for property
				CodeMethodReturnStatement getter = new CodeMethodReturnStatement(new CodeVariableReferenceExpression("_" + typeToWrap.Name + "." + fieldinfo.Name));
				prop.GetStatements.Add(getter);

				//create setter for property
				CodeAssignStatement setter = new CodeAssignStatement(new CodeVariableReferenceExpression("_" + typeToWrap.Name + "." + fieldinfo.Name), new CodeVariableReferenceExpression("value"));
				prop.SetStatements.Add(setter);

				//add the property to the class
				code.Members.Add(prop);
			}

			//create and add namespace
			CodeNamespace ns = new CodeNamespace(typeToWrap.Name + "Wrapper");
			ns.Types.Add(code);

			//create codeunit and add the namespace
			CodeCompileUnit codeUnit = new CodeCompileUnit();
			codeUnit.Namespaces.Add(ns);

			//set compiler parameters
			CompilerParameters compilerParams = new CompilerParameters();
			compilerParams.GenerateInMemory = true;
			String assemblyFileName = System.Reflection.Assembly.GetExecutingAssembly().GetName().CodeBase;
			if(assemblyFileName.IndexOf("file:///") > -1)
				compilerParams.ReferencedAssemblies.Add(assemblyFileName.Substring("file:///".Length));
			
			//Compile the codeunit
			ICodeCompiler compiler = new Microsoft.CSharp.CSharpCodeProvider().CreateCompiler();
			CompilerResults results = compiler.CompileAssemblyFromDom(compilerParams, codeUnit);
			if(results.Errors.HasErrors)
			{
				StringBuilder errorString = new StringBuilder("compilation errors:\n");
				CompilerError err;
				IEnumerator it = results.Errors.GetEnumerator();
				while(it.MoveNext())
				{
					err = (CompilerError)it.Current;
					errorString.Append(err.ToString() + "\n");
				}
				throw new ApplicationException(errorString.ToString());
			}	

			return results.CompiledAssembly.GetType(typeToWrap.Name + "Wrapper." + typeToWrap.Name + "WithProperties");
		}

		public static ArrayList GetArrayList(IList arrayToConvert)
		{
			lock(typeof(swall.WebServiceWrapper))
			{
				if(arrayToConvert.Count > 0)
				{
					//add instances of the new class to a new ArrayList
					Object oldObj;
					Type newType = GetWrapperType(arrayToConvert[0].GetType());
					ArrayList newArray = new ArrayList();
					IEnumerator it = arrayToConvert.GetEnumerator();
					while(it.MoveNext())
					{
						oldObj = it.Current;
						Object newObj = Activator.CreateInstance(newType);
						System.Reflection.MethodInfo mi = newType.GetMethod("SetInnerObject");
						object[] args = new object[] { oldObj };
						mi.Invoke(newObj,System.Reflection.BindingFlags.Default | System.Reflection.BindingFlags.InvokeMethod, null, args, null);
						newArray.Add(newObj);
					}
					return newArray;
				}
				else
				{
					return null;
				}
			}
		}
	}
}
